//-------------------------------------------------------------------//
//                                                                   //
//  WInfo Copyright (C) Glenn Rosenthal, 2000.  All rights reserved. //
//                                                                   //   
//  This utility program processes information from several soruces  //
//  to generate kana->kanji conversion dictionaries used by JWPce.   //
//                                                                   //
//  The main input data files come from the WNN package which was    //
//  deleveloped by Kyoto Universinity, OMRON, and ASTEC.  Please     //
//  See the copyright notice below.                                  //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------//
//
//  Copyright notice for WNN files (most of the .u files)       
//
//    Copyright Kyoto University Research Institute for Mathematical 
//            Sciences 1987, 1988, 1989, 1990, 1991
//    Copyright OMRON Corporation. 1987, 1988, 1989, 1990, 1991
//    Copyright ASTEC, Inc. 1987, 1988, 1989, 1990, 1991
//
//    Permission to use, copy, modify, distribute, and sell this software and 
//    its documentation for any purpose is hereby granted without fee, provided 
//    that all of the following conditions are satisfied:
//
//        1. The above copyright notices appear in all copies
//        2. Both those copyright notices and this permission notice appear 
//            in supporting documentation
//        3. The name of "Wnn" isn't changed unless substantial modifications 
//            are made, or the following words followed by the above copyright 
//            notices appear in all supporting documentation of software based 
//            on "Wnn":
//
//            "This software is based on the original version of Wnn developed 
//            by Kyoto University Research Institute for Mathematical Sciences 
//            (KURIMS), OMRON Corporation and ASTEC Inc."
//
//        4. The names KURIMS, OMRON and ASTEC may not be used in advertising 
//            or publicity pertaining to distribution of the software without 
//            specific, written prior permission of KURIMS, OMRON and ASTEC, 
//            who make no representations about the suitability of this software 
//            for any purpose. It is provided "as is" without express or implied 
//            warranty.
//
//  Wnn consortium is one of distributors of the official Wnn source code 
//  release. Wnn consortium also makes no representations about the suitability 
//  of this software for any purpose. It is provided "as is" without express or 
//  implied warranty.
//
//  KURIMS, OMRON, ASTEC and Wnn consortium disclaim all warranties with regard 
//  to this software, including all implied warranties of merchantability and 
//  fitness. In no event shall KURIMS, OMRON, ASTEC or Wnn consortium be liable 
//  for any special, indirect or consequential damages or any damages whatsoever 
//  resulting from loss of use, data or profits, whether in an action of contract, 
//  negligence or other tortuous action, arising out of or in connection with the 
//  use or performance of this software.
//
//-------------------------------------------------------------------//

typedef unsigned short ushort;      // Convenient definitions.
typedef unsigned char  byte;
typedef unsigned short KANJI;       

#define BASE_KATAKANA   0x2500      // Base value for katakana characters.
#define BASE_HIRAGANA   0x2400      // Base value for hiragana characters.
#define BASE_JASCII     0x2300      // Page for JASCII
#define BASE_KANJI      0x3000      // Value of first actual kanji

#define KANJI_SPACE     0x2121      // Space character in kanji.
#define KANJI_SLASH     0x213F      // Forward slash (Japanese).
#define KANJI_CAMA      0x2122      // Cama (Japanese)
#define KANJI_DOT       0x2126      // Center dot (Japanese)
#define KANJI_ASTRIC    0x2176      // Astric (Japanese)
#define KANJI_ARROW     0x222A      // Horizontal arrow (use in user conversion editor)
#define KANJI_TILDE     0x2141      // ~ (Japanese)
#define KANJI_DASH      0x213D      // - (Japanese, not long vawel, just a dash)
#define KANJI_LONGVOWEL 0x213C      // - (Japanese)
#define KANJI_LPARAN    0x214A      // ( (Japanese)
#define KANJI_RPARAN    0x214B      // ) (Japanese)
#define KANJI_LBRACKET  0x215A      // [ (heavy Japanese)
#define KANJI_RBRACKET  0x215B      // ] (heavy Japanese)   
#define KANJI_BAD       0x2223      // Knaji used for invalid kanji (this is a solid box).

#define KANJI_U         0x2426
#define KANJI_KU        0x242F
#define KANJI_GU        0x2430
#define KANJI_SU        0x2439
#define KANJI_ZU        0x243A
#define KANJI_TU        0x2444
#define KANJI_DU        0x2445
#define KANJI_NU        0x244C
#define KANJI_FU        0x2455
#define KANJI_BU        0x2456
#define KANJI_PU        0x2457
#define KANJI_MU        0x2460
#define KANJI_RU        0x246B
#define KANJI_I         0x2424
#define KANJI_TA        0x243F

#define ISCRLF(ch)     (((ch) == '\n') || ((ch) == '\r'))               // Is this a cr or lf 
#define ISSPACE(ch)    ((ch == ' ') || (ch == '\t') || (ch == 0x2121))  // Is this a space.
#define ISASCII(ch)    (((ch) & 0x7f00) == 0)                           // Is this any ASCII character
#define ISJIS(ch)      (((ch) & 0x7f00) != 0)                           // Is this a kanji or kana
#define ISHIRAGANA(ch) (((ch) & 0x7f00) == BASE_HIRAGANA)               // Is this hiragana
#define ISKATAKANA(ch) (((ch) & 0x7f00) == BASE_KATAKANA)               // Is this katakana
#define ISKANA(ch)     (ISHIRAGANA(ch) || ISKATAKANA(ch))               // Is this a kana
#define ISJASCII(ch)   (((ch) & 0x7f00) == BASE_JASCII)
#define ISKANJI(ch)    (((ch)         ) >= BASE_KANJI)

                                // Return values from WNN_node::match()
#define MATCH_NONE  0           // Does not match.
#define MATCH_PART  1           // Kana and godan match.
#define MATCH_ALL   2           // Full match.

//-------------------------------------------------------------------//
//                                                                   //
//  KANJI_index class.                                               //
//                                                                   //
//  This class defines an entry in the WNN index file.               //
//                                                                   //
//-------------------------------------------------------------------//

typedef class KANJI_index {
public:
  byte key[4];                  // Index search key
  long offset;                  // Offset in the dictionary
  void clear (void);            // Clear the structure
} KANJI_index;

//-------------------------------------------------------------------//
//                                                                   //
//  WNN_node class.                                                  //
//                                                                   //
//  This class defines a node.  Initally, this is a single entry     //
//  into the conversion list.  After the merge, this may represent   //
//  several possible entries.                                        //
//                                                                   //
//-------------------------------------------------------------------//

#define SIZE_KANJI  400                     // Length of buffer for node (need to handle merge)
#define SIZE_KANA   50                      // Size of the kana and key buffers

#define TYPE_NOUN       1                   // Anyting other than a godan verb.
#define TYPE_GODAN      2                   // Godan verb, has special endings.
#define TYPE_ICHIDAN    3                   // Ichidan verb, allows more flexable endings.
#define TYPE_IADJ       4                   // i-adjitive, allows more flexable endings.
#define TYPE_DELETE     100                 // Marker, delete this entry.
#define TYPE_SPECIAL    101                 // Returned for irregular gozaru->gozaimas godan verbs.
#define TYPE_ERROR      102                 // Error marker.

#define GODAN_NONE  0

typedef class WNN_node {
public:
  WNN_node       (void);                            // Blank constructor.
  WNN_node       (byte *buffer);                    // New node from a WNN data file.
  WNN_node       (byte *buffer,int priority);       // New node from DIC data file.
  WNN_node       (class WNN_node *node,KANJI end);  // Generate a node from other node
  int  add       (KANJI k);                         // Append a knaji to the kanji buffer.
  void debug     (FILE *file);                      // Print debuggin inforamtion for this node.
  int  match     (class WNN_node *node);            // Do two modes match (are they identical conversions)
  int  merge     (class WNN_node *node);            // Merge another node at the end of this one.
  int  mtest     (class WNN_node *node);            // Test for mergablitlity (same key, same godan).
  int  test      (class WNN_node *node);            // Determine relative order of two nodes.
  int  test_key  (KANJI_index *index);              // Test if key has changed.        
  int  write     (FILE *file);                      // Write a dictionary file.
  int  write_new (FILE *file);                      // Write node to dictionary in new format
  int  dic_verb  (KANJI end,int _godan,int _type,byte *meaning);  // Format a verb from dictionary format
  int  dic_two   (KANJI end1,KANJI end2,byte *meaning);           // Remove two and make a noun.

  int   type;                                       // Type of node (realy just TYPE_GODAN/TYPE_NOUN/TYPE_DELETE).
private:
  KANJI kana [SIZE_KANA ];                          // Kana for this node (not used after read)
  KANJI kanji[SIZE_KANJI];                          // Kanji for this node.
  byte  key  [SIZE_KANA ];                          // Key for node -- Key is generated from the low byte of the kana.
  int   wnn_pri;                                    // WNN priority -- High is higher priority
  int   kanji_pri;                                  // Kanji priority -- Low is higher priority
  int   godan;                                      // Godan pattern (u,r,s,...).  A value of '1' indicates an ichidan verb.
  int   len;                                        // Length of the kanji string.
} WNN_node;

typedef class WNN_index {
public:
  WNN_index           (void);
  ~WNN_index          (void);
  int       init      (long size);                  // Allocate index array.
  int       debug     (char *name);                 // Write debugging file with state of index list.
  void      filter    (void);                       // Filter duplicate entries.
  WNN_node *find      (WNN_node *node);             // Find a specific node if it exists.
  int       insert    (WNN_node *node);             // Insert node into list.
  int       irregular (WNN_node *node);             // Insert routine to handle irregular verbs (gozaru->gozaimasu)
  int       merge     (void);                       // Scans list and merges enties that can be merged.
  int       read_dic  (char *name,int priority);    // Read a dictionary file.
  int       read_line (char *buffer);               // Adds a single line entry to the list.
  int       read_wnn  (char *name);                 // Reads WNN tyle .u files
  void      sort      (void);                       // Sort list.
  int       write     (char *base);                 // Write older JWP dictionary and index file.
  int       write_new (char *base);                 // Write newer JWPce dictionary and index file.
  long      count;                                  // Number of nodes
private:
  void      quicksort (long low,long high);         // Actuall sorting routine.
  class WNN_node **list;                            // Actual list.
  long             alloc;                           // Number of allocated spaces
} WNN_index;




